# -*- coding: utf-8 -*-

import os
import logging
import logging.handlers

from colorlog import ColoredFormatter

CONSOLE_LEVEL = 1000


def setup_logging(name):
    """
    Setup initial logging configuration
    """

    assert isinstance(name, str)
    
    # Add console level
    logging.addLevelName(CONSOLE_LEVEL, "CONSOLE_LEVEL")

    def console(self, message, *args, **kws):  # pragma no cover
        # Yes, logger takes its '*args' as 'args'.
        if self.isEnabledFor(CONSOLE_LEVEL):
            self._log(CONSOLE_LEVEL, message, args, **kws)

    logging.Logger.console = console

    # Init logger
    logger = logging.getLogger(name)

    # Set file log format
    file_format = logging.Formatter('[%(levelname)s] %(asctime)s - %(message)s', "%Y-%m-%d %H:%M:%S")
    log_file = logging.FileHandler(filename=os.path.join(os.getcwd(), "aiotasks.log"))

    log_file.setFormatter(file_format)

    # Handler: console
    formatter = ColoredFormatter(
        "[ %(log_color)s*%(reset)s ] %(blue)s%(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'white',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'red,bg_white',
            'CONSOLE_LEVEL': 'green'
        },
        secondary_log_colors={},
        style='%'
    )
    log_console = logging.StreamHandler()
    log_console.setFormatter(formatter)

    # --------------------------------------------------------------------------
    # Add all of handlers to logger config
    # --------------------------------------------------------------------------
    logger.addHandler(log_console)
    logger.addHandler(log_file)

__all__ = ("setup_logging", "CONSOLE_LEVEL")



import os
import atexit
import asyncio
import logging
from typing import Union

try:
    import ujson as json
except ImportError:
    import json
    

from .bases import AsyncTaskBase, AsyncTaskSubscribeBase, AsyncTaskDelayBase
from .redis import AsyncTaskDelayRedis, AsyncTaskSubscribeRedis
from .memory import AsyncTaskSubscribeMemory, AsyncTaskDelayMemory

log = logging.getLogger("aiotasks")


if not os.getenv("AIOTASK_DEBUG", False):  # pragma: no cover
    try:
        import uvloop
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        
        log.debug("uvloop found. Using it as the event loop")
    except ImportError:
        pass
    

# --------------------------------------------------------------------------
# Composed backends
# --------------------------------------------------------------------------
class MemoryBackend(AsyncTaskDelayMemory, AsyncTaskSubscribeMemory, AsyncTaskBase):
    
    def __init__(self, prefix: str = "aoitasks", loop=None):  # pragma: no cover
        self.loop = loop or asyncio.get_event_loop()
        self.prefix = prefix

        AsyncTaskSubscribeMemory.__init__(self, loop=loop)
        AsyncTaskDelayMemory.__init__(self, prefix=prefix, loop=loop)
        AsyncTaskBase.__init__(self)

        # This line is necessary to close redis connections
        atexit.register(self.stop)


class RedisBackend(AsyncTaskSubscribeRedis, AsyncTaskDelayRedis, AsyncTaskBase):
    
    def __init__(self, dsn, prefix: str = "aiotasks", loop=None):
        self.loop = loop or asyncio.get_event_loop()

        AsyncTaskSubscribeRedis.__init__(self, dsn=dsn, prefix=prefix, loop=loop)
        AsyncTaskDelayRedis.__init__(self, dsn=dsn, prefix=prefix, loop=loop)
        AsyncTaskBase.__init__(self)

        # This line is necessary to close redis connections
        atexit.register(self.stop)
        

def build_manager(dsn: str = "memory://",
                  prefix: str = "aiotasks",
                  loop=None) -> Union[AsyncTaskBase, AsyncTaskSubscribeBase, AsyncTaskDelayBase]:
    
    loop = loop or asyncio.get_event_loop()
    
    if not prefix:
        log.error("Empty task_prefix. Using 'aiotasks' as task_prefix")
        prefix = "aiotasks"
    
    # Fixing task_prefix type
    prefix = str(prefix)

    if not os.getenv("AIOTASK_DEBUG", False):  # pragma: no cover
        loop.set_debug(True)
    
    if dsn.startswith("memory"):
        ret = MemoryBackend(prefix=prefix, loop=loop)
    elif dsn.startswith("redis"):
        ret = RedisBackend(dsn=dsn, prefix=prefix, loop=loop)

    # Store manager for global access
    import builtins
    builtins.__aiotasks__ = ret
    
    return ret

__all__ = ("build_manager", )



import asyncio

from aiotasks import build_manager


loop = asyncio.get_event_loop()
loop.set_debug(True)

manager = build_manager(loop=loop)


@manager.task()
async def task_01(num):
    print("Task 01 starting: {}".format(num))
    
    await asyncio.sleep(2, loop=loop)
    
    print("Task 01 stopping")
    
    return "a"


async def main_async():
    manager.run()
    
    async with task_01.wait(1) as f:
        print(f)
    
    await manager.wait(5)
    
    manager.stop()


if __name__ == '__main__':
    loop.run_until_complete(main_async())



from aiotasks import build_manager


def test_build_manager_invalid_prefix(event_loop):
    import logging
    
    logger = logging.getLogger("aiotasks")
    
    class CustomLogger(logging.StreamHandler):
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []
        
        def emit(self, record):
            self.content.append(record.msg)
    
    custom = CustomLogger()
    logger.addHandler(custom)
    
    manager = build_manager(dsn="memory://", loop=event_loop, prefix=None)
    
    assert manager.task_prefix == "aiotasks"
    assert "Empty task_prefix. Using 'aiotasks' as task_prefix" in custom.content


def test_build_manager_invalid_prefix_type(event_loop):
    manager = build_manager(dsn="memory://", loop=event_loop, prefix=11)
    
    assert manager.task_prefix == "11"



#!/usr/bin/env python3

import click
import logging

from aiotasks import global_options

from .console import launch_aiotasks_in_console


log = logging.getLogger('aiotasks')


@global_options()
@click.pass_context
def cli(ctx, **kwargs):  # pragma no cover
    ctx.obj = kwargs


@cli.command(help="Launch aiotasks")
@click.pass_context
@click.option('-A', '--app', 'application', required=True)
@click.option('--loglevel', '-l', 'log_level')
@click.option('--config', 'config_file')
def worker(ctx, **kwargs):
    launch_aiotasks_in_console(ctx.obj, **kwargs)


if __name__ == "__main__" and __package__ is None:  # pragma no cover
    cli()



from aiotasks import parse_dsn


def test_parse_dsn_runs_ok():
    dsn = "redis://root:root@127.0.0.1"
    
    user, password, host, port, path = parse_dsn(dsn)
    
    assert user == "root"
    assert password == "root"
    assert host == "127.0.0.1"
    assert port is None
    assert path == ""


def test_parse_dsn_user_and_no_password():
    dsn = "redis://root@127.0.0.1:9999"
    
    user, password, host, port, path = parse_dsn(dsn)
    
    assert user == "root"
    assert password == ""
    assert host == "127.0.0.1"
    assert port == 9999
    assert path == ""


def test_parse_dsn_empty_password():
    dsn = "redis://root:@127.0.0.1:9999"
    
    user, password, host, port, path = parse_dsn(dsn)
    
    assert user == "root"
    assert password == ""
    assert host == "127.0.0.1"
    assert port == 9999
    assert path == ""


def test_parse_dsn_empty_valid_path():
    dsn = "redis://root:@127.0.0.1:9999/20"
    
    user, password, host, port, path = parse_dsn(dsn)
    
    assert user == "root"
    assert password == ""
    assert host == "127.0.0.1"
    assert port == 9999
    assert path == "20"


def test_parse_dsn_empty_long_path():
    dsn = "redis://root:@127.0.0.1:9999/20/aaa/bbb"
    
    user, password, host, port, path = parse_dsn(dsn)
    
    assert user == "root"
    assert password == ""
    assert host == "127.0.0.1"
    assert port == 9999
    assert path == "20/aaa/bbb"



import time
import uuid
import asyncio

import logging
import msgpack

from aiotasks import build_manager

log = logging.getLogger("aiotasks")


# --------------------------------------------------------------------------
# Testing @task decorator
# --------------------------------------------------------------------------
def test_redis_delay_task_decorator_oks(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    globals()["test_redis_delay_task_decorator_oks_finished"] = False
    
    @manager.task()
    async def task_test_redis_delay_task_decorator_oks(num):
        globals()["test_redis_delay_task_decorator_oks_finished"] = True
    
    async def run():
        manager.run()
        
        await task_test_redis_delay_task_decorator_oks.delay(1)
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_delay_task_decorator_oks_finished"] is True
    
    del globals()["test_redis_delay_task_decorator_oks_finished"]


def test_redis_delay_task_decorator_no_port_gotten(event_loop, redis_instance):
    
    _redis_instance, _ = redis_instance.rsplit(":", maxsplit=1)
    
    manager = build_manager(dsn=_redis_instance, loop=event_loop)
    
    globals()["test_redis_delay_task_decorator_no_port_gotten_finished"] = False
    
    @manager.task()
    async def task_test_redis_delay_task_decorator_oks_no_port_gotten(num):
        globals()["test_redis_delay_task_decorator_no_port_gotten_finished"] = True
    
    async def run():
        manager.run()
        
        await task_test_redis_delay_task_decorator_oks_no_port_gotten.delay(1)
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_delay_task_decorator_no_port_gotten_finished"] is True
    
    del globals()["test_redis_delay_task_decorator_no_port_gotten_finished"]


def test_redis_delay_task_decorator_timeout_raises(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    globals()["test_redis_delay_task_decorator_timeout_raises_finished_tasks"] = False
    
    @manager.task()
    async def task_test_redis_delay_task_decorator_timeout_raises(num):
        await asyncio.sleep(num, loop=event_loop)
        globals()["test_redis_delay_task_decorator_timeout_raises_finished_tasks"] = True
    
    async def run():
        manager.run()
        
        await task_test_redis_delay_task_decorator_timeout_raises.delay(1)
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_delay_task_decorator_timeout_raises_finished_tasks"] is False
    
    del globals()["test_redis_delay_task_decorator_timeout_raises_finished_tasks"]


def test_redis_delay_task_decorator_check_correct_timeout_reached(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    _start = time.time()
    
    @manager.task()
    async def task_test_redis_delay_task_decorator_check_correct_timeout_reached(num):
        pass
    
    async def run():
        manager.run()
        
        await task_test_redis_delay_task_decorator_check_correct_timeout_reached.delay(4)
        
        await manager.wait(timeout=0.5, exit_on_finish=False, wait_timeout=0.2)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    _stop = time.time()
    
    _running_time = _stop - _start
    
    assert _running_time > 0.5


def test_redis_delay_task_decorator_invalid_function(event_loop, redis_instance):
    
    import logging
    
    logger = logging.getLogger("aiotasks")
    
    class CustomLogger(logging.StreamHandler):
        
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []
        
        def emit(self, record):
            self.content.append(record.msg)
    
    custom = CustomLogger()
    logger.addHandler(custom)
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    async def run():
        
        # Send an invalid task name
        task_id = uuid.uuid4().hex
        
        await manager._redis_poller.lpush(manager.task_list_name,
                                          msgpack.packb(dict(task_id=task_id,
                                                             function="non_exist",
                                                             args=(),
                                                             kwargs={})))
        
        manager.run()
        
        await manager.wait(timeout=0.2, exit_on_finish=False, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert "No local task with name 'non_exist'" in custom.content


def test_redis_delay_task_decorator_invalid_task_id_format(event_loop, redis_instance):
    
    import random
    import logging
    
    logger = logging.getLogger("aiotasks")
    
    class CustomLogger(logging.StreamHandler):
        
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []
        
        def emit(self, record):
            self.content.append(record.msg)
    
    custom = CustomLogger()
    logger.addHandler(custom)
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    task_id = random.randint(10, 1000)
    
    @manager.task()
    async def task_test_redis_delay_task_decorator_invalid_task_id_format():
        pass
    
    async def run():
        manager.run()
        
        # Send an invalid task name
        await manager._redis_poller.lpush(manager.task_list_name,
                                          msgpack.packb(dict(task_id=task_id,
                                                             function="task_test_redis_delay_task_decorator_invalid_task_id_format",
                                                             args=(),
                                                             kwargs={})))
        
        await manager.wait(timeout=0.5, exit_on_finish=False, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert "Task ID '{}' has not valid UUID4 format".format(task_id) in custom.content


def test_redis_delay_task_decorator_custom_task_name(event_loop, redis_instance):
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    @manager.task(name="custom_test_redis_delay_task_decorator_custom_name")
    async def task_test_redis_delay_task_decorator_custom_name():
        pass
    
    async def run():
        manager.run()
        
        await manager.wait(timeout=0.5, exit_on_finish=False, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert "custom_test_redis_delay_task_decorator_custom_name" in manager.task_available_tasks.keys()


# --------------------------------------------------------------------------
# Testing add_task
# --------------------------------------------------------------------------
def test_redis_delay_add_task_oks(event_loop, redis_instance):
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    globals()["test_task_test_redis_delay_add_task_oks_finished"] = False
    
    async def task_test_redis_delay_add_task_oks(num):
        globals()["test_task_test_redis_delay_add_task_oks_finished"] = True
    
    async def run():
        manager.run()
        
        # Add task without decorator
        manager.add_task(task_test_redis_delay_add_task_oks)
        
        await task_test_redis_delay_add_task_oks.delay(1)
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_task_test_redis_delay_add_task_oks_finished"] is True
    
    del globals()["test_task_test_redis_delay_add_task_oks_finished"]


def test_redis_delay_add_task_custom_task_name(event_loop, redis_instance):
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    async def task_test_redis_delay_add_task_custom_task_name():
        pass
    
    async def run():
        manager.run()
        
        # Add task without decorator
        manager.add_task(task_test_redis_delay_add_task_custom_task_name,
                         name="custom_task_test_redis_delay_task_decorator_oks")
        
        await task_test_redis_delay_add_task_custom_task_name.delay()
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()

    assert "custom_task_test_redis_delay_task_decorator_oks" in manager.task_available_tasks.keys()


def test_redis_delay_add_task_non_coroutine_as_input(event_loop, redis_instance):
    import logging
    
    logger = logging.getLogger("aiotasks")
    
    class CustomLogger(logging.StreamHandler):
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []
        
        def emit(self, record):
            self.content.append(record.msg)
    
    custom = CustomLogger()
    logger.addHandler(custom)
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    def task_test_redis_delay_add_task_non_coroutine_as_input():
        pass
    
    async def run():
        manager.run()
        
        # Add task without decorator
        manager.add_task(task_test_redis_delay_add_task_non_coroutine_as_input,
                         name="custom_task_test_redis_delay_task_decorator_oks")
        
    event_loop.run_until_complete(run())
    manager.stop()

    assert "Function 'task_test_redis_delay_add_task_non_coroutine_as_input' is not a coroutine and can't be added as a task" in custom.content

    # assert "custom_task_test_redis_delay_task_decorator_oks" in manager._tasks.keys()



# from aiotasks import build_manager
#
#
# def test_redis_decorated_functions_has_correct_dynamic_functions(event_loop, redis_instance):
#     manager = build_manager(dsn=redis_instance, loop=event_loop)
#
#     globals()["has_delay_method"] = False
#     globals()["has_wait_method"] = False
#     globals()["has_subscribe_method"] = False
#
#     @manager.task()
#     async def task_test_redis_decorated_functions_has_correct_dynamic_functions():
#         pass
#
#     async def run():
#         manager.run()
#
#         globals()["has_delay_method"] = hasattr(task_test_redis_decorated_functions_has_correct_dynamic_functions,
#                                                 "delay")
#
#         await manager.wait(timeout=0.5, exit_on_finish=True, wait_timeout=0.2)
#
#     event_loop.run_until_complete(run())
#     manager.stop()
#
#     assert globals()["has_delay_method"] is True
#
#     del globals()["has_delay_method"]
#



import time
import uuid
import asyncio

import msgpack

from aiotasks import build_manager

from subapp import *

manager = build_manager(dsn="redis://127.0.0.1")


@manager.task()
async def task_test_redis_delay_oks(num):
    globals()["test_redis_delay_oks_finished"] = True


async def run():
    manager.run()
    
    await task_test_redis_delay_oks.delay(1)
    
    await manager.wait(timeout=1, exit_on_finish=True, wait_timeout=0.2)


if __name__ == '__main__':
    manager.loop.run_until_complete(run())
    
    hola()
    manager.stop()


from aiotasks import build_manager


def test_redis_decorated_functions_has_correct_dynamic_functions(event_loop, redis_instance):
    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["has_delay_method"] = False
    
    @manager.task()
    async def task_test_redis_decorated_functions_has_correct_dynamic_functions():
        pass
    
    async def run():
        manager.run()

        globals()["has_delay_method"] = hasattr(task_test_redis_decorated_functions_has_correct_dynamic_functions,
                                                "delay")
        
        await manager.wait(timeout=0.5, exit_on_finish=True, wait_timeout=0.2)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["has_delay_method"] is True
    
    del globals()["has_delay_method"]




import random
import asyncio

from aiotasks import build_manager


loop = asyncio.get_event_loop()
loop.set_debug(True)

manager = build_manager(dsn="redis://127.0.0.1/0", loop=loop)


@manager.task()
async def task_01(num):
    # wait_time = random.randint(1,)
    wait_time = 0.01
    
    print("Task {}. Waiting for {} seconds".format(num, wait_time))
    
    await asyncio.sleep(wait_time, loop=loop)
    
    print("Task {} stopping".format(num))


async def main_async():
    for i in range(1000):
        print("T{}".format(i))
        await task_01.delay(random.randint(1, 100))

if __name__ == '__main__':
    loop.run_until_complete(main_async())



import uuid
import asyncio
import logging

try:
    import umsgpack as msgpack
except ImportError:  # pragma: no cover
    import msgpack

from .bases import *
from .context import AsyncWaitContextManager

log = logging.getLogger("aiotasks")


class MemoryAsyncWaitContextManager(AsyncWaitContextManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    def __await__(self, *args, **kwargs):
        task_id = uuid.uuid4().hex
        
        return asyncio.ensure_future(self.poller.put((self.list_name,
                                                      msgpack.packb(dict(task_id=task_id,
                                                                         function=self.fn.function_name,
                                                                         args=self.args,
                                                                         kwargs=self.kwargs),
                                                                    use_bin_type=True))),
                                     loop=self.loop).__await__()


class AsyncTaskSubscribeMemory(AsyncTaskSubscribeBase):
    
    def __init__(self,
                 prefix: str = "aiotasks",
                 loop=None):
        super().__init__(prefix, loop)
        
        self._loop_subscribers = loop or asyncio.get_event_loop()
        self.topics_messages = asyncio.Queue(loop=self.loop_subscribers)
    
    async def publish(self, topic, info):
        await self.topics_messages.put(("{}:{}".format(self.prefix, topic),
                                        msgpack.packb(dict(topic=topic,
                                                           data=info),
                                                      use_bin_type=True)))
    
    async def has_pending_topics(self):
        return not self.topics_messages.empty()

    async def register_topics(self):
        pass

    async def wait_for_message(self, channel):
        return True

    async def get_next_message(self, channel):
        return await self.topics_messages.get()

    def stop_subscriptions(self):
        for t in self.running_tasks.values():
            t.cancel()
    

# --------------------------------------------------------------------------
# Delayers
# --------------------------------------------------------------------------
class AsyncTaskDelayMemory(AsyncTaskDelayBase):
    
    def __init__(self,
                 dsn=None,
                 prefix: str = "aiotasks",
                 loop=None,
                 concurrency: int = 5):
        super().__init__(prefix, loop, concurrency)
        
        self._task_queue = asyncio.Queue(loop=self.loop_delay)
    
    async def has_pending_tasks(self):
        return not self._task_queue.empty() or bool(self.task_running_tasks)
    
    def custom_task_done(self, task_id):
        self._task_queue.task_done()
    
    def stop_delayers(self):
        for t in self.task_running_tasks.values():
            t.cancel()
    
    @property
    def pending_tasks(self):
        return self._task_queue.get()
    
    @property
    def context_class(self):
        return MemoryAsyncWaitContextManager
    
    @property
    def poller(self):
        return self._task_queue


__all__ = ("AsyncTaskSubscribeMemory", "AsyncTaskDelayMemory")



import asyncio
import os
import importlib

from functools import partial

from aiotasks import AsyncTaskBase
from .model import *

# here = os.path.abspath(os.path.dirname(__file__))
here = os.getcwd()
get_path = partial(os.path.join, here)


def run_default_aiotasks(config: AioTasksDefaultModel):
    assert isinstance(config, AioTasksDefaultModel)
    
    _module = str(config.application)
    
    if _module.endswith(".py"):
        _module = _module.replace(".py", "")
    
    # Try to find AsyncTaskBase var
    app = importlib.import_module(_module)
    
    manager = None
    for v in dir(app):
        if v.startswith("_"):
            continue
        
        value = getattr(app, v)
        
        if isinstance(value, AsyncTaskBase):
            manager = value
            break
            
    if not manager:
        raise AttributeError("Not found AsyncTaskBase subclass instance")
    
    manager.run()
    try:
        manager.loop.run_forever()
    finally:
        for task in asyncio.Task.all_tasks(loop=manager.loop):
            task.cancel()
        manager.loop.run_forever()

__all__ = ("run_default_aiotasks", )



import pytest

from aiotasks.core.helpers import dict_to_obj


def test_dict_to_obj_response_ok():
    
    ret = dict_to_obj(dict(hello="world", bye="see you"))
    
    assert hasattr(ret, "hello")
    assert hasattr(ret, "bye")


def test_dict_to_obj_response_invalid_input():
    
    with pytest.raises(AssertionError):
        dict_to_obj(None)


def test_dict_to_obj_response_empty():
    
    assert issubclass(dict_to_obj({}), object)



def main():
    import os
    import sys
    
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    sys.path.insert(1, parent_dir)
    import aiotasks
    
    __package__ = str("aiotasks")
    
    # Run the cmd
    from aiotasks.actions.default.cli import cli
    
    cli()

if __name__ == "__main__":  # pragma no cover
    main()




"""
This file contains utils and reusable functions
"""

from collections import namedtuple


def dict_to_obj(data):
    """
    Transform an input dict into a object.

    >>> data = dict(hello="world", bye="see you")
    >>> obj = dict_to_obj(data)
    >>> obj.hello
    'world'

    :param data: input dictionary data
    :type data: dict
    """
    assert isinstance(data, dict)
    
    if not data:
        return namedtuple("OBJ", [])
    
    obj = namedtuple("OBJ", list(data.keys()))
    
    return obj(**data)

__all__ = ("dict_to_obj", )




from aiotasks import SharedConfig, String


class AioTasksDefaultModel(SharedConfig):
    application = String()
    log_level = String()
    config_file = String()
    
__all__ = ("AioTasksDefaultModel", )





import pytest
import asyncio

from aiotasks import build_manager, AioTasksTimeout


def test_memory_wait_oks(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_wait_oks_finished"] = False

    @manager.task()
    async def task_test_memory_wait_oks():
        return True

    async def run():
        manager.run()

        async with task_test_memory_wait_oks.delay() as f:
            globals()["test_memory_wait_oks_finished"] = f

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_memory_wait_oks_finished"] == True

    del globals()["test_memory_wait_oks_finished"]


def test_memory_wait_timeout_raises(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_wait_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_memory_wait_oks():
        await asyncio.sleep(2, loop=event_loop)

        return True

    async def run():
        manager.run()

        try:
            async with task_test_memory_wait_oks.delay(timeout=0.2) as f:
                pass
        except AioTasksTimeout:
                globals()["test_memory_wait_timeout_raises_finished"] = True

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_memory_wait_timeout_raises_finished"] is True

    del globals()["test_memory_wait_timeout_raises_finished"]


def test_memory_wait_infinite_timeout_raises(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_wait_infinite_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_memory_wait_oks():
        await asyncio.sleep(2, loop=event_loop)

        return True

    async def run():
        manager.run()
        try:
            async with task_test_memory_wait_oks.delay(infinite_timeout=0.2) as f:
                globals()["test_memory_wait_infinite_timeout_raises_finished"] = f
        except AioTasksTimeout:
            globals()["test_memory_wait_timeout_raises_finished"] = True

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_memory_wait_infinite_timeout_raises_finished"] == False

    del globals()["test_memory_wait_infinite_timeout_raises_finished"]


def test_memory_wait_infinite_raises_timeout_exception(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_wait_infinite_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_memory_wait_oks():
        await asyncio.sleep(2, loop=event_loop)

        return True

    async def run():
        manager.run()
        with pytest.raises(AioTasksTimeout):
            async with task_test_memory_wait_oks.delay(infinite_timeout=0.2) as f:
                pass

    event_loop.run_until_complete(run())
    manager.stop()




import re
import os
import click
import codecs

#
# Get version software version
#
version_file = os.path.join(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")), '__init__.py')
with codecs.open(version_file, 'r', 'latin1') as fp:  # pragma no cover
    try:
        version = re.findall(r"^__version__ = ['\"]([^']+)['\"]\r?$",
                             fp.read(), re.M)[0]
    except IndexError:
        raise RuntimeError('Unable to determine version.')
    
# --------------------------------------------------------------------------
# Common options for command line interface
# --------------------------------------------------------------------------
global_options_list = (
    # General
    click.option('-v', 'verbosity', count=True, type=int, default=1, help='Verbose output'),
    click.option('--quiet', '-q', 'verbosity', flag_value=0, help='Minimal output'),
    click.version_option(version=version)
)


class global_options(object):
    def __init__(self, invoke_without_command=False):
        assert isinstance(invoke_without_command, bool)
        
        self.invoke_without_command = invoke_without_command
    
    def __call__(self, f):
        def wrapped_f(*args):
            fn = f
            for option in reversed(global_options_list):
                fn = option(f)
            
            fn = click.group(context_settings={'help_option_names': ['-h', '--help']},
                             invoke_without_command=self.invoke_without_command)(fn)
            
            return fn
        
        return wrapped_f()


#
# HERE MORE EXAMPLES OF CMD OPTIONS
#
# --------------------------------------------------------------------------
# Options for "auto" command
# --------------------------------------------------------------------------
#
# auto_options_list = (
#     click.option('-T', '--timeout', 'timeout', type=int, default=60,
#                  help="max time to wait until actions are available"),
# )
#
#
# class auto_options(object):
#     def __call__(self, f):
#         def wrapped_f(*args):
#             fn = f
#             for option in reversed(auto_options_list):
#                 fn = option(f)
#
#             return fn
#
#         return wrapped_f()


__all__ = ("global_options",)  # "auto_options")



# -*- coding: utf-8 -*-

from booby import *


class SharedConfig(Model):
    verbosity = Integer(default=0)
    # worker = String(choices=['process', 'threads'])
    # task_concurrency = Integer()
    timeout = Integer(default=10)



import pytest

from aiotasks.core.logger import setup_logging


def test_setup_logging_runs_ok():
    assert setup_logging("blah") is None


def test_setup_logging_runs_null_as_name():
    
    with pytest.raises(AssertionError):
        setup_logging(None)


def test_setup_logging_runs_invalid_as_name():
    with pytest.raises(AssertionError):
        setup_logging(dict())



import time
import uuid
import asyncio

import msgpack

from aiotasks import build_manager


def test_memory_subscribers_oks(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_subscribers_oks_finished"] = False

    @manager.subscribe("hello")
    async def task_test_memory_subscribers_oks(topic, data):

        if topic == "hello" and data == "world":
            globals()["test_memory_subscribers_oks_finished"] = True

    async def run():
        manager.run()

        await manager.publish("hello", "world")

        await manager.wait(timeout=0.2, wait_timeout=0.1)

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_memory_subscribers_oks_finished"] is True

    del globals()["test_memory_subscribers_oks_finished"]


def test_memory_subscribers_no_topics(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    async def run():
        manager.run()

        await manager.publish("hello", "world")

        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.2)

    event_loop.run_until_complete(run())
    manager.stop()

    assert len(manager.topics_subscribers) == 0


def test_memory_subscribers_empty_topics(event_loop):
    import logging

    logger = logging.getLogger("aiotasks")

    class CustomLogger(logging.StreamHandler):
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []

        def emit(self, record):
            self.content.append(record.msg)

    custom = CustomLogger()
    logger.addHandler(custom)

    manager = build_manager(dsn="memory://", loop=event_loop)

    @manager.subscribe()
    async def task_test_memory_subscribers_oks(topic, data):
        if topic == "hello" and data == "world":
            globals()["test_memory_subscribers_oks_finished"] = True

    async def run():
        manager.run()

        await manager.publish("hello", "world")

        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.2)

    event_loop.run_until_complete(run())
    manager.stop()

    assert len(manager.topics_subscribers) == 0
    assert "Empty topic fount in function 'task_test_memory_subscribers_oks'. Skipping it." in custom.content


def test_memory_subscribers_duplicated_topics(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    @manager.subscribe("hello")
    async def task_test_memory_subscribers_oks(topic, data):
        pass

    @manager.subscribe("hello")
    async def task_test_memory_subscribers_oks_2(topic, data):
        pass

    async def run():
        manager.run()

        await manager.publish("hello", "world")

        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)

    event_loop.run_until_complete(run())
    manager.stop()

    assert len(manager.topics_subscribers) == 1


def test_memory_subscribers_timeout_raises(event_loop):

    manager = build_manager(dsn="memory://", loop=event_loop)

    globals()["test_memory_subscribers_timeout_raises_finished_tasks"] = False

    @manager.subscribe("hello")
    async def task_test_memory_subscribers_oks(topic, data):
        if topic == "hello":
            await asyncio.sleep(data, loop=event_loop)
            globals()["test_memory_subscribers_timeout_raises_finished_tasks"] = True

    async def run():
        manager.run()

        await manager.publish("hello", 5)

        await manager.wait(timeout=0.5, exit_on_finish=True, wait_timeout=0.1)

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_memory_subscribers_timeout_raises_finished_tasks"] is False

    del globals()["test_memory_subscribers_timeout_raises_finished_tasks"]





# import click
#
# from click.testing import CliRunner
#
# from aiotasks.actions.default.cli import info
#
# import aiotasks.actions.default.console
#
#
# def _launch_aiotasks_in_console(blah, **kwargs):
#     click.echo("ok")
#
#
# def test_cli_info_runs_show_help():
#     runner = CliRunner()
#     result = runner.invoke(info)
#
#     assert 'Usage: info [OPTIONS] ' in result.output
#
#
# def test_cli_info_runs_ok():
#     # Patch the launch of: launch_aiotasks_info_in_console
#     aiotasks.actions.default.cli.launch_aiotasks_in_console = _launch_aiotasks_in_console
#
#     runner = CliRunner()
#     result = runner.invoke(info, ["aaaa"])
#
#     assert 'ok' in result.output



import sys
import pytest

from aiotasks.__main__ import main


def test_aiotasks___main__runs_ok():
    # This test checks that the main command line run well
    
    sys.argv = [sys.argv[0], "-h"]
    
    with pytest.raises(SystemExit) as e:
        main()
        
    assert str(e.value) == '0'



from urllib.parse import urlparse


def parse_dsn(dsn: str) -> tuple:
    """Get a DSN and return pased info as:

    >>> dsn="redis://user:password@127.0.0.1:6379/0"
    >>> parse_dsn(dsn)
    "user", "password", "127.0.0.1", 6379, "0"
    """
    _, host, db, _, _, _ = urlparse(dsn)
    
    # Get user / Pass
    if "@" in host:
        credentials, host = host.split("@", maxsplit=1)
    else:
        credentials = None

    user, password = None, None
    if credentials:
        _splitted_cred = credentials.split(":", maxsplit=1)
        
        if len(_splitted_cred) == 1:
            user = _splitted_cred[0]
            password = ""
        else:
            user, password = _splitted_cred
    
    if ":" in host:
        host, port = host.split(":")
        port = int(port)
    else:
        port = None
    
    if db.startswith("/"):
        db = db[1:]
    if not db:
        db = ""
    
    return user, password, host, port, db


__all__ = ("parse_dsn", )



import pytest
import asyncio

from aiotasks import build_manager, AioTasksTimeout


def test_redis_wait_oks(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["test_redis_wait_oks_finished"] = False
    
    @manager.task()
    async def task_test_redis_wait_oks():
        return True
    
    async def run():
        manager.run()

        async with task_test_redis_wait_oks.delay() as f:
            globals()["test_redis_wait_oks_finished"] = f
        
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_wait_oks_finished"] == True
    
    del globals()["test_redis_wait_oks_finished"]
    

def test_redis_wait_no_port_gotten(event_loop, redis_instance):
    
    _redis_instance, _ = redis_instance.rsplit(":", maxsplit=1)
    
    manager = build_manager(dsn=_redis_instance, loop=event_loop)

    globals()["test_redis_wait_no_port_gotten_finished"] = False

    @manager.task()
    async def task_test_redis_wait_no_port_gotten():
        return True

    async def run():
        manager.run()
    
        async with task_test_redis_wait_no_port_gotten.delay() as f:
            globals()["test_redis_wait_no_port_gotten_finished"] = f
    
    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_redis_wait_no_port_gotten_finished"] == True
    
    del globals()["test_redis_wait_no_port_gotten_finished"]


def test_redis_wait_timeout_raises(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["test_redis_wait_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_redis_wait_oks():
        await asyncio.sleep(2, loop=event_loop)
        
        return True

    async def run():
        manager.run()

        try:
            async with task_test_redis_wait_oks.delay(timeout=0.2) as f:
                pass
        except AioTasksTimeout:
                globals()["test_redis_wait_timeout_raises_finished"] = True
            
    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_redis_wait_timeout_raises_finished"] is True
    
    del globals()["test_redis_wait_timeout_raises_finished"]


def test_redis_wait_infinite_timeout_raises(event_loop, redis_instance):

    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["test_redis_wait_infinite_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_redis_wait_oks():
        await asyncio.sleep(2, loop=event_loop)

        return True

    async def run():
        manager.run()
        try:
            async with task_test_redis_wait_oks.delay(infinite_timeout=0.2) as f:
                globals()["test_redis_wait_infinite_timeout_raises_finished"] = f
        except AioTasksTimeout:
            globals()["test_redis_wait_timeout_raises_finished"] = True

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_redis_wait_infinite_timeout_raises_finished"] == False

    del globals()["test_redis_wait_infinite_timeout_raises_finished"]


def test_redis_wait_infinite_raises_timeout_exception(event_loop, redis_instance):

    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["test_redis_wait_infinite_timeout_raises_finished"] = False

    @manager.task()
    async def task_test_redis_wait_oks():
        await asyncio.sleep(2, loop=event_loop)

        return True

    async def run():
        manager.run()
        with pytest.raises(AioTasksTimeout):
            async with task_test_redis_wait_oks.delay(infinite_timeout=0.2) as f:
                pass

    event_loop.run_until_complete(run())
    manager.stop()




import concurrent


class AioTasksError(Exception):
    pass


class AioTasksValueError(ValueError):
    pass


class AioTasksTypeError(TypeError):
    pass


class AioTasksTimeout(concurrent.futures.TimeoutError):
    pass


__all__ = ("AioTasksError", "AioTasksValueError", "AioTasksTypeError", "AioTasksTimeout")



import time
import socket
import inspect
import os.path
import subprocess

import redis
import pytest
import redislite


@pytest.fixture(scope="session")
def redis_instance():
    """
    This fixture ensures that a Redis instance is launched and yield the DSN to connecto to it
    """
    
    host = "127.0.0.1"
    port = "6379"
    
    # Is an Redis server launched?
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    if sock.connect_ex((host, int(port))) == 0:
        yield "redis://{}:{}".format(host, port)
    
    # NO Redis server found. Launch and instance
    else:
        redis_bin = os.path.join(os.path.dirname(inspect.getfile(redislite)), "bin", "redis-server")
        
        # Launch Redis server
        p = subprocess.Popen([redis_bin,
                              "--port", port,
                              "--bind", host])
        
        r = redis.Redis(host=host, port=int(port))
        
        # Wait until redis is available
        while True:
            if r.ping():
                break
            time.sleep(0.5)
        
        del r
        
        yield "redis://{}:{}".format(host, port)

        p.kill()



import abc
import uuid
import asyncio
import logging
import concurrent


try:
    import umsgpack as msgpack
except ImportError:  # pragma: no cover
    import msgpack
    
from ..core.exceptions import AioTasksTimeout


log = logging.getLogger("aiotasks")


class AsyncWaitContextManager:
    
    def __init__(self,
                 *args,
                 **kwargs):
        self.fn = args[0]
        self.list_name = args[1]
        self.poller = args[2]
        self.loop = args[3] or asyncio.get_event_loop()
        self.timeout = kwargs.pop("timeout", 0)
        self.infinite_timeout = kwargs.pop("infinite_timeout", 900)
        
        self.args = args[4:]
        self.kwargs = kwargs
    
    @abc.abstractmethod
    def __await__(self, *args, **kwargs):  # pragma: no cover
        pass

    async def __aenter__(self):
        # Timeout != 0 -> apply timeout
        try:
            if self.timeout:
                return await asyncio.wait_for(self.fn(*self.args, **self.kwargs),
                                              timeout=self.timeout,
                                              loop=self.loop)
            # Timeout == 0 -> infinite --> Apply very long timeout
            else:
                return await asyncio.wait_for(self.fn(*self.args, **self.kwargs),
                                              timeout=self.infinite_timeout,
                                              loop=self.loop)
        
        except concurrent.futures.TimeoutError as e:
            log.error("{function}: {error_message}".format(function=self.fn.__name__,
                                                           error_message=e))
            raise AioTasksTimeout(e) from e
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass
    
    
__all__ = ("AsyncWaitContextManager", )



# -*- coding: utf-8 -*-

import uuid
import logging
import asyncio
import aioredis

from functools import partial
from collections import defaultdict

from .bases import *
from ..helpers import parse_dsn
from .context import AsyncWaitContextManager

try:
    import umsgpack as msgpack
except ImportError:  # pragma: no cover
    import msgpack

log = logging.getLogger("aiotasks")


class RedisAsyncWaitContextManager(AsyncWaitContextManager):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __await__(self, *args, **kwargs):
        task_id = uuid.uuid4().hex
        
        return self.poller.lpush(self.list_name,
                                 msgpack.packb(dict(task_id=task_id,
                                                    function=self.fn.function_name,
                                                    args=self.args,
                                                    kwargs=self.kwargs),
                                               use_bin_type=True)).__await__()
    

# --------------------------------------------------------------------------
# Subscribers
# --------------------------------------------------------------------------
class AsyncTaskSubscribeRedis(AsyncTaskSubscribeBase):
    
    def __init__(self,
                 dsn: str = "redis://127.0.0.1:6379/0",
                 prefix: str = "aiotasks",
                 loop=None):
        super().__init__(prefix, loop)
        
        _, password, host, port, db = parse_dsn(dsn)
        
        if not port:
            port = 6379
        
        port = int(port)
        if not db:
            db = 0
        
        self._redis_pub = self.loop_subscribers.run_until_complete(aioredis.create_redis(address=(host, port),
                                                                                         db=db,
                                                                                         password=password,
                                                                                         loop=self.loop_subscribers))
        
        self._redis_sub = self.loop_subscribers.run_until_complete(aioredis.create_redis(address=(host, port),
                                                                                         db=db,
                                                                                         password=password,
                                                                                         loop=self.loop_subscribers))
        
    async def publish(self, topic, info):
        # Wait for channel listener
        if not self.subscriber_ready.is_set():
            await self.subscriber_ready.wait()
        
        await self._redis_pub.publish("{}:{}".format(self.prefix, topic),
                                      msgpack.packb(dict(topic=topic, data=info),
                                                    use_bin_type=True))
    
    async def has_pending_topics(self):
        return bool(self.running_tasks)
    
    async def register_topics(self):
        channels = await self._redis_sub.psubscribe("{}:*".format(self.prefix))
        return channels[0]
    
    async def wait_for_message(self, channel):
        return await channel.wait_message()
        
    async def get_next_message(self, channel):
        return await channel.get()
        
    def stop_subscriptions(self):
        self._redis_sub.close()
        self._redis_pub.close()
        
        if not self.loop_subscribers.is_closed():
            self.loop_subscribers.run_until_complete(self._redis_sub.wait_closed())
            self.loop_subscribers.run_until_complete(self._redis_pub.wait_closed())
    

# --------------------------------------------------------------------------
# Delayers
# --------------------------------------------------------------------------
class AsyncTaskDelayRedis(AsyncTaskDelayBase):
    
    def __init__(self,
                 dsn: str = "redis://127.0.0.1:6379/0",
                 prefix: str = "aiotasks",
                 concurrency: int = 5,
                 loop=None):
        super().__init__(prefix, loop, concurrency)
        
        _, password, host, port, db = parse_dsn(dsn)
        
        if not port:
            port = 6379
        
        port = int(port)
        if not db:
            db = 0
        
        self._redis_consumer = self.loop_delay.run_until_complete(aioredis.create_redis(address=(host, port),
                                                                                        db=db,
                                                                                        password=password,
                                                                                        loop=self.loop_delay))
        
        self._redis_poller = self.loop_delay.run_until_complete(aioredis.create_redis(address=(host, port),
                                                                                      db=db,
                                                                                      password=password,
                                                                                      loop=self.loop_delay))
    
    async def has_pending_tasks(self):
        return bool(self.task_running_tasks) or not bool(await self._redis_poller.llen(self.task_list_name))
        
    def stop_delayers(self):
        self._redis_consumer.close()
        self._redis_poller.close()
        
        for t in self.task_running_tasks.values():
            t.cancel()
        
        if not self.loop_delay.is_closed():
            self.loop_delay.run_until_complete(self._redis_consumer.wait_closed())
            self.loop_delay.run_until_complete(self._redis_poller.wait_closed())
    
    @property
    def poller(self):
        return self._redis_poller
    
    @property
    def context_class(self):
        return RedisAsyncWaitContextManager
    
    @property
    def pending_tasks(self):
        return self._redis_consumer.brpop(self.task_list_name, timeout=0)
    

__all__ = ("AsyncTaskSubscribeRedis", "AsyncTaskDelayRedis")



import uvloop
import asyncio

from aiotasks import build_manager

asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())

loop = asyncio.get_event_loop()
loop.set_debug(True)

manager = build_manager(loop=loop)


@manager.subscribe("hola")
async def task_01(topic, data):
    print("Task subscribe task_01 - ({}): {}".format(topic, data))
    
    await asyncio.sleep(2, loop=loop)
    
    print("Task 01 stopping")


@manager.subscribe(["hola", "mundo"])
async def task_02(topic, data):
    print("Task subscribe task_02 - ({}): {}".format(topic, data))
    
    await asyncio.sleep(2, loop=loop)
    
    print("Task 01 stopping")


async def main_async():
    manager.run()
    
    for x in range(10000):
        await manager.publish("mundo", "daaaata{}".format(x))
    await manager.wait(10)
    
    manager.stop()


if __name__ == '__main__':
    loop.run_until_complete(main_async())



import random
import asyncio

from aiotasks import build_manager

event_loop = asyncio.new_event_loop()

manager = build_manager(dsn="redis://127.0.0.1/1", loop=event_loop)

globals()["test_redis_delay_oks_finished"] = False


@manager.task()
async def task_test_redis_delay_oks(num):
    await asyncio.sleep(0.1, loop=event_loop)
    print("#" * 50)
    globals()["test_redis_delay_oks_finished"] = True


async def run():
    manager.run()
    
    # await task_test_redis_delay_oks.delay(1)

    for x in range(10):
        print("POOLING")
        t = await manager._redis_poller.strlen(manager._list_name)
        
        print("RESULT: ", t)
        
        await asyncio.sleep(0.5, loop=event_loop)
    
    # await manager.wait(timeout=5, exit_on_finish=True, wait_timeout=0.5)


if __name__ == '__main__':
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_delay_oks_finished"] is True
    
    del globals()["test_redis_delay_oks_finished"]



import abc
import time
import uuid
import asyncio
import logging

from functools import partial
from collections import defaultdict

try:
    import umsgpack as msgpack
except ImportError:  # pragma: no cover
    import msgpack

log = logging.getLogger("aiotasks")


# --------------------------------------------------------------------------
# Base classes
# --------------------------------------------------------------------------
class AsyncTaskSubscribeBase(metaclass=abc.ABCMeta):
    
    def __init__(self,
                 prefix: str = "aiotasks",
                 loop=None):
        self.loop_subscribers = loop or asyncio.get_event_loop()
        self.prefix = prefix

        self.running_tasks = dict()
        self.topics_subscribers = defaultdict(set)
        self.subscriber_ready = asyncio.Event(loop=self.loop_subscribers)
    
    def subscribe(self, topics=None):
        """Decorator"""
    
        if not topics:
            topics = set()
    
        if isinstance(topics, str):
            topics = {topics}
    
        def real_decorator(f):
            # if function is a coro, add some new functions
            if asyncio.iscoroutinefunction(f):
                if not topics:
                    log.error("Empty topic fount in function '{}'. Skipping it.".format(f.__name__))
                for topic in topics:
                    self.topics_subscribers[topic].add(f)
            return f
    
        return real_decorator
    
    @abc.abstractmethod
    async def wait_for_message(self, channel):  # pragma: no cover
        pass
    
    @abc.abstractmethod
    async def get_next_message(self, channel) -> object:  # pragma: no cover
        pass
    
    @abc.abstractmethod
    async def register_topics(self):  # pragma: no cover
        pass
    
    @abc.abstractmethod
    async def publish(self, topic, info):  # pragma: no cover
        pass
    
    @abc.abstractmethod
    async def has_pending_topics(self):  # pragma: no cover
        pass

    def _make_tasks_done_subscriber(self, task_id, future):
        self.running_tasks.pop(task_id)

    async def listen_topics(self):
        # Mark ready OK
        self.subscriber_ready.set()
    
        channel = await self.register_topics()
    
        while await self.wait_for_message(channel):
            raw = await self.get_next_message(channel)
        
            if hasattr(raw, "__iter__") and len(raw) != 2:
                log.error("Invalid data from Redis subscriber. It must be a tuple with len 2")
        
            ch, data = raw
        
            if hasattr(ch, 'decode'):
                ch = ch.decode()
        
            # Get topic
            try:
                prefix, topic = ch.split(":", maxsplit=1)
            except ValueError:
                log.error("Invalid channel name: {}".format(ch))
                continue
        
            # Check prefix
            if prefix != self.prefix:
                log.error("Invalid prefix: {}".format(prefix))
                continue
                
            if hasattr(data, "encode"):
                data = data.encode()
        
            msg = msgpack.unpackb(data, encoding='utf-8')
            data_topic = msg.get("topic", False)
            data_content = msg.get("data", False)
        
            if not data_topic or not data_content:
                log.error("Invalid data topic / data content - topic: {} / data: {}".format(data_topic, data_content))
                continue
        
            for fn in self.topics_subscribers.get(topic, tuple()):
                # Build stop task
                task_id = uuid.uuid4().hex
                done_fn = partial(self._make_tasks_done_subscriber, task_id)
            
                task = self.loop_subscribers.create_task(fn(data_topic, data_content))
                task.add_done_callback(done_fn)
            
                self.running_tasks[task_id] = task

    @abc.abstractmethod
    def stop_subscriptions(self):  # pragma: no cover
        pass
    

class AsyncTaskDelayBase(metaclass=abc.ABCMeta):
    
    # --------------------------------------------------------------------------
    # Implemented methods
    # --------------------------------------------------------------------------
    def __init__(self,
                 prefix: str = "aiotasks",
                 loop=None,
                 concurrency: int = 5):
        
        self.task_prefix = prefix
        self.task_running_tasks = dict()
        self.task_available_tasks = dict()
        self.task_concurrency = concurrency
        self.loop_delay = loop or asyncio.new_event_loop()
        self.task_list_name = "{}:{}".format(self.task_prefix, "tasks")
        
        # Semaphore for task_concurrency
        self.task_concurrency_sem = asyncio.BoundedSemaphore(self.task_concurrency,
                                                             loop=self.loop_delay)

    def task(self, name: str = None):
        """Decorator"""
    
        def real_decorator(f):
            # Real call to funcion
            def new_f(*args, **kwargs):
                return f(*args, **kwargs)
        
            # if function is a coro, add some new functions
            if asyncio.iscoroutinefunction(f):
                new_f.delay = partial(self.context_class,
                                      new_f,
                                      self.task_list_name,
                                      self.poller,
                                      self.loop_delay)
            
                if name:
                    function_name = name
                else:
                    function_name = f.__name__
            
                new_f.function_name = function_name
            
                self.task_available_tasks[function_name] = f
        
            return new_f
    
        return real_decorator

    def add_task(self, function: callable, name: str = None) -> callable:
        if not asyncio.iscoroutinefunction(function):
            log.warning("Function '{}' is not a coroutine and can't be added as a task".format(function.__name__))
            return
    
        function.delay = partial(self.context_class,
                                 function,
                                 self.task_list_name,
                                 self.poller,
                                 self.loop_delay)
    
        if name:
            function_name = name
        else:
            function_name = function.__name__
    
        function.function_name = function_name
    
        self.task_available_tasks[function_name] = function

    def _make_tasks_done_delay(self, running_task, future):
        self.task_running_tasks.pop(running_task)
        self.task_concurrency_sem.release()
    
        if hasattr(self, "custom_task_done"):
            self.custom_task_done(running_task)

    async def listen_tasks(self):
        while True:
            raw_data = await self.pending_tasks
        
            # Limit concurrent executions
            await self.task_concurrency_sem.acquire()
        
            _, raw = raw_data
        
            msg = msgpack.unpackb(raw, encoding='utf-8')
            args = msg.get("args")
            kwargs = msg.get("kwargs")
            task_id = msg.get("task_id")
            function = msg.get("function")
        
            try:
                if type(task_id) is int:
                    task_id = str(task_id)
            
                uuid.UUID(task_id, version=4)
            except ValueError:
                log.error("Task ID '{}' has not valid UUID4 format".format(task_id))
                continue
        
            try:
                local_task = self.task_available_tasks[function]
            except KeyError:
                log.warning("No local task with name '{}'".format(function))
                continue
        
            running_task_id = uuid.uuid4().hex
            done_fn = partial(self._make_tasks_done_delay, running_task_id)
        
            # Build stop task
            task = self.loop_delay.create_task(local_task(*args, **kwargs))
            task.add_done_callback(done_fn)
        
            self.task_running_tasks[running_task_id] = task

    # --------------------------------------------------------------------------
    # Abstract methods & properties
    # --------------------------------------------------------------------------
    @abc.abstractmethod
    async def poller(self):  # pragma: no cover
        pass
    
    @abc.abstractmethod
    async def has_pending_tasks(self):  # pragma: no cover
        pass
    
    @abc.abstractmethod
    def stop_delayers(self):  # pragma: no cover
        pass

    # --------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------
    @abc.abstractproperty
    def pending_tasks(self):  # pragma: no cover
        pass
    
    @abc.abstractproperty
    def context_class(self):  # pragma: no cover
        pass


class AsyncTaskBase(object, metaclass=abc.ABCMeta):
    def __init__(self):
        self._launcher_tasks = None
        self._launcher_topics = None
    
    async def wait(self, *, timeout: float = 0, exit_on_finish: bool = False, wait_timeout: float = 1.0):
        """
        :param exit_on_finish: exit when all pending tasks are finished
        :type exit_on_finish: bool
        
        :param timeout: Time in seconds
        :type timeout: int
        """
        TIME_STEP = wait_timeout
        
        _infinite = False
        if timeout == 0:
            _infinite = True
        _start_time = time.time()
        
        while True:
            _has_pending_tasks = await self.has_pending_tasks()
            _has_pending_topics = await self.has_pending_topics()
            
            # There are pending tasks?
            if _has_pending_tasks or _has_pending_topics:
                # Yes, there's pending tasks, but is timeout reached?
                if time.time() - _start_time > timeout and not _infinite:
                    return
            else:
                
                # No tasks pending and marked -> If marked as a exit on tasks finished
                if exit_on_finish:
                    return
                
                # NO, there's not pending tasks, but is timeout reached?
                if time.time() - _start_time > timeout and not _infinite:
                    return
            
            # Wait
            await asyncio.sleep(TIME_STEP, loop=self.loop)
    
    def stop(self):
        self.stop_delayers()
        self.stop_subscriptions()
        
        if self._launcher_topics:
            self._launcher_topics.cancel()
        if self._launcher_tasks:
            self._launcher_tasks.cancel()

        if not self.loop_delay.is_closed():
            self.loop_delay.run_until_complete(asyncio.wait([self._launcher_tasks], loop=self.loop_delay))
        if not self.loop_subscribers.is_closed():
            self.loop_subscribers.run_until_complete(asyncio.wait([self._launcher_topics], loop=self.loop_subscribers))

    def run(self):
        """Blocking run"""
        self._launcher_topics = self.loop_subscribers.create_task(self.listen_topics())
        self._launcher_tasks = self.loop_delay.create_task(self.listen_tasks())


__all__ = ("AsyncTaskSubscribeBase", "AsyncTaskDelayBase", "AsyncTaskBase")



import time
import uuid
import asyncio

import msgpack

from aiotasks import build_manager


def test_redis_subscribers_oks(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    globals()["test_redis_subscribers_oks_finished"] = False
    
    @manager.subscribe("hello")
    async def task_test_redis_subscribers_oks(topic, data):
        
        if topic == "hello" and data == "world":
            globals()["test_redis_subscribers_oks_finished"] = True
    
    async def run():
        manager.run()
        
        await manager.publish("hello", "world")
        
        await manager.wait(timeout=0.5, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert globals()["test_redis_subscribers_oks_finished"] is True
    
    del globals()["test_redis_subscribers_oks_finished"]


def test_redis_subscribers_no_topics(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    async def run():
        manager.run()
        
        await manager.publish("hello", "world")
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.2)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert len(manager.topics_subscribers) == 0
    
    
def test_redis_subscribers_empty_topics(event_loop, redis_instance):
    import logging
    
    logger = logging.getLogger("aiotasks")
    
    class CustomLogger(logging.StreamHandler):
        def __init__(self):
            super(CustomLogger, self).__init__()
            self.content = []
        
        def emit(self, record):
            self.content.append(record.msg)
    
    custom = CustomLogger()
    logger.addHandler(custom)
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)

    @manager.subscribe()
    async def task_test_redis_subscribers_oks(topic, data):
        if topic == "hello" and data == "world":
            globals()["test_redis_subscribers_oks_finished"] = True
    
    async def run():
        manager.run()
        
        await manager.publish("hello", "world")
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.2)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert len(manager.topics_subscribers) == 0
    assert "Empty topic fount in function 'task_test_redis_subscribers_oks'. Skipping it." in custom.content


def test_redis_subscribers_duplicated_topics(event_loop, redis_instance):
    
    manager = build_manager(dsn=redis_instance, loop=event_loop)
    
    @manager.subscribe("hello")
    async def task_test_redis_subscribers_oks(topic, data):
        pass
    
    @manager.subscribe("hello")
    async def task_test_redis_subscribers_oks_2(topic, data):
        pass
    
    async def run():
        manager.run()
        
        await manager.publish("hello", "world")
        
        await manager.wait(timeout=0.2, exit_on_finish=True, wait_timeout=0.1)
    
    event_loop.run_until_complete(run())
    manager.stop()
    
    assert len(manager.topics_subscribers) == 1


def test_redis_subscribers_no_port_gotten(event_loop, redis_instance):

    _redis_instance, _ = redis_instance.rsplit(":", maxsplit=1)

    manager = build_manager(dsn=_redis_instance, loop=event_loop)

    globals()["test_redis_subscribers_oks_finished"] = False

    @manager.subscribe("hello")
    async def task_test_redis_subscribers_oks(topic, data):
        if topic == "hello" and data == "world":
            globals()["test_redis_subscribers_oks_finished"] = True

    async def run():
        manager.run()

        await manager.publish("hello", "world")

        await manager.wait(timeout=0.5, exit_on_finish=True, wait_timeout=0.2)

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_redis_subscribers_oks_finished"] is True

    del globals()["test_redis_subscribers_oks_finished"]


def test_redis_subscribers_timeout_raises(event_loop, redis_instance):

    manager = build_manager(dsn=redis_instance, loop=event_loop)

    globals()["test_redis_subscribers_timeout_raises_finished_tasks"] = False

    @manager.subscribe("hello")
    async def task_test_redis_subscribers_oks(topic, data):
        if topic == "hello":
            await asyncio.sleep(data, loop=event_loop)
            globals()["test_redis_subscribers_timeout_raises_finished_tasks"] = True

    async def run():
        manager.run()

        await manager.publish("hello", 5)

        await manager.wait(timeout=0.5, exit_on_finish=True, wait_timeout=0.1)

    event_loop.run_until_complete(run())
    manager.stop()

    assert globals()["test_redis_subscribers_timeout_raises_finished_tasks"] is False

    del globals()["test_redis_subscribers_timeout_raises_finished_tasks"]





import logging

from .model import *
from ..helpers import *
from .api import run_default_aiotasks

log = logging.getLogger('aiotasks')


def launch_aiotasks_in_console(shared_config, **kwargs):
    """Launch in console mode"""
    
    # Load config
    config = AioTasksDefaultModel(**shared_config, **kwargs)
    
    # Check if config is valid
    if not config.is_valid:
        for prop, msg in config.validation_errors:
            log.critical("[!] '%s' property %s" % (prop, msg))
        return
    
    log.setLevel(config.verbosity)
    
    try:
        log.console("Starting aioTasks")
        
        run_default_aiotasks(config)
    
    except KeyboardInterrupt:
        log.console("[*] CTRL+C caught. Exiting...")
    except Exception as e:
        log.critical("[!] Unhandled exception: %s" % str(e))
        
        log.exception("[!] Unhandled exception: %s" % e, stack_info=True)
    finally:
        log.debug("[*] Shutdown...")
        

__all__ = ("launch_aiotasks_in_console",)
